;******************************************************************************
;*                                                                            *
;*                            License and T&C                                 *
;*                                                                            *
;*        GR8BIT storage subsystem microcontroller microcode version 1.0      *
;*             This software is copyleft 2011 by Eugeny Brychkov              *
;*                                                                            *
;* You can use and modify its source and executable, given two conditions are *
;* met: (a) you keep this Copyleft statement in the source code, and you do   *
;* not distribute software and resulting executable on the commercial basis   *
;* and on form the purposes, not related to GR8BIT platform, without written  *
;* permission of Eugeny Brychkov (eugeny_brychkov@hotmail.com)                *
;*                                                                            *
;******************************************************************************

; version 1.0 (Nov 28,2011)		- initial release
; version 1.0a (Jan 26,2012)	- Storage subsystem ROM was released, this code was intact
; version 1.0b (Mar 14, 2012)	- added support for 720K media, fixed some bugs regarding
;								  handling of i/o errors and disk removals, improved
;								  disk change handling. No retry algorithm is implemented
;								  so far. If r/w operation fails and both speeds are tried,
;								  execution is terminated
; version 2.0 (Apr 02,2012)		- initial version for PIC18F46K22, almost entirely discarded
;								  handshaking alorithm, added VERSION command

;///////////////////////////////////////////////////////////////////////////////
;
;							TECHNICAL INFORMATION
; GR8BIT storage subsystem based on PIC18F46K22 does not require additional
; wait states at the 3.579545MHz CPU speed
;
; The principle of operation of the storage subsystem
; There're two ports in GR8BIT, &hA4 (port base+0) and &hA5 (port base+1), which
; control storage subsystem microcontroller. Port base+0 is used for command
; initiation and abortion (in some cases it is not possible to abort command in
; execution). Port base+1 is used for light-weight handshaking and data transfers.
;
; Microcontroller is running at 64MHz system clock (62.5ns instruction cycle)
;
;///////////////////////////////////////////////////////////////////////////////

; GAP lengths can be found in SMC37C65 data sheet
; Info on DSPEC values can be found here http://wiki.osdev.org/Floppy_Disk_Controller, also http://www.isdaman.com/alsos/hardware/fdc/floppy.htm
; and also http://www.mathematik.uni-kl.de/~hannes/en/floppy.diff

; 1.44MB diskette / speed configuration
GPL1_144	equ		0x1B		; gap length when reading
GPL2_144	equ		0x54		; gap length when formatting
DSPEC1_144	equ		0xAF		; A=step rate 6ms, F=head unload delay (240ms)
DSPEC2_144	equ		0x11		; 8=16ms head load time, 1=no DMA
DSPD_144	equ		0x0			; speed 500kB/s
SpT_144		equ		0x12		; sectors per track (18)

; 720K diskette / speed configuration
GPL1_720	equ	0x2A
GPL2_720	equ	0x50
DSPEC1_720	equ	0xDF		; D=step rate 6ms, F=head unload time (480ms)
DSPEC2_720	equ	0x11		; 8=32ms head load time (very safe), 1=no DMA
DSPD_720	equ	0x02		; speed 250kB/s (CR1=1, CR0=0, DRV=x)
SpT_720		equ	0x09		; sectors per track (9)

	list		p=18f46k22			; list directive to define processor
	#include	<P18F46K22.inc>		; processor specific variable definitions

; Microcontroller configuration
	CONFIG	FOSC=INTIO67		; internal oscillator, RA6 and RA7 are generic I/O pins
	CONFIG	PLLCFG=ON			; turn PLL (4*) on
	CONFIG	FCMEN=OFF			; fail-safe clock monitor off
	CONFIG	IESO=OFF			; internal/external oscillator switching is off
	CONFIG	PWRTEN=OFF			; power up timer off (using internal oscillator only)
	CONFIG	BOREN=OFF			; brown-out reset disabled completely
	CONFIG	WDTEN=OFF			; watchdog timer is controlled by SWDTEN bit
	CONFIG	PBADEN=OFF			; port B bits 4:0 are digital I/O
	CONFIG	MCLRE=INTMCLR		; MCRLE function is off (pin is used as digital input)
	CONFIG	LVP=OFF
	CONFIG	STVREN=OFF

; Command buffer memory organization
;	output: ACCESS bank, sequence of the bytes (or words) to send to controller (offset 0x00-0x0F)
;	intput: ACCESS bank, sequence of the bytes (or words) read from the controller (offset 0x10-0x1F)
CMDWLOC		equ	0x00			; location of the queue which is written to the FDC
CMDRLOC		equ	0x10			; location of the bytes which are read from FDC

; Data buffer memory organization: starts at address 0x100 (addressed using indirect addressing scheme)

; System variables
CTRSTAT		equ	0x20			; controller status - free/busy
TMP			equ	0x21
MAINSTAT	equ	0x22			; last read master status register of FDD
SENSEST		equ	0x23			; last read drive sense status (status reg 3 - ST3)
FDCQPOS		equ	0x24
QTEMP		equ	0x25
CMDRUN		equ	0x26			; not zero if command is running
FDCRUNPOS	equ	0x27			; position of the pointer when writing sequence to the FDD controller
RESULTN		equ	0x28			; number of result bytes to read from FDD

FMTTMP1		equ	0x29
FMTTMP2		equ	0x2A
FMTTMP3		equ	0x2B

DEBUGN		equ	0x2C

TIMOUT1		equ	0x2D			; used within timout counting code
TIMOUT2		equ	0x2E			; used within timout counting code
FDCERR		equ	0x2F			; not 0 if error
DATAVAL		equ	0x30
MOTATO		equ	0x31			; motor drive A timeout (0=no timeout)
MOTBTO		equ	0x32			; motor drive B timeout (0=no timeout)
DORVAL		equ	0x33			; current value of the DOR
CURDRV		equ	0x34			; Current disk drive (0=A, 1=B)

RECLENL		equ	0x35
RECLENH		equ	0x36

;DSKCHGFLG	equ	0x37			; disk change flag saved to be returned during next DSKCHG call
HOSTCMD		equ	0x38
TEMPREQ		equ	0x39			; temporary value for forming request to FDC
TEMPRE1		equ	0x3A			; temporary value for forming request to FDC
COUVAL		equ	0x3B			; command/MC-host interface counter value
DRINUM		equ	0x3C			; number of physical drives connected to the FDC
DRVPRM		equ	0x3D			; drive and head parameter in FDC format (CURDRV is just drive number 0 or 1)
CYLPRM		equ	0x3E			; cylinder parameter in FDC format
SECPRM		equ	0x3F			; sector parameter in FDC format
FMTPRM		equ	0x40
FMTPR1		equ	0x41			; three variables for format's writing control records
FMTPR2		equ	0x42
FMTPR3		equ	0x43
CURHOSTDRV	equ	0x44

SPDCOU		equ	0x45			; Counter for speed retry
DSK0SPD		equ	0x46			; DSK*SPD contain speed current disk/disk drive is using
DSK1SPD		equ	0x47			; 0=500kB/s, 1=250kB/s
CURGPL1		equ	0x48
CURGPL2		equ	0x49
CURDSP1		equ	0x4A
CURDSP2		equ	0x4B
CURDSPD		equ	0x4C
CURSPT		equ	0x4D
DSK0CHGFLG	equ	0x4E			; disk change flags for drives 0 and 1
DSK1CHGFLG	equ	0x4F

DELAY1		equ	0x5D	; used within delay counting code
DELAY2		equ	0x5E	; used within delay counting code
DELAY3		equ	0x5F	; used within delay counting code
; --- Last variable space offset within ACCESS bank = 0x5F


; PROGRAM START
	org	0
	goto	main

; INTERRUPT VECTOR SETTING
	org	8
	goto	int_handler

Set144SpeedVals:
	movlw	GPL1_144
	movwf	CURGPL1,ACCESS
	movlw	GPL2_144
	movwf	CURGPL2,ACCESS
	movlw	DSPEC1_144
	movwf	CURDSP1,ACCESS
	movlw	DSPEC2_144
	movwf	CURDSP2,ACCESS
	movlw	DSPD_144
	movwf	CURDSPD,ACCESS
	movlw	SpT_144
	movwf	CURSPT,ACCESS
	return

Set720SpeedVals:
	movlw	GPL1_720
	movwf	CURGPL1,ACCESS
	movlw	GPL2_720
	movwf	CURGPL2,ACCESS
	movlw	DSPEC1_720
	movwf	CURDSP1,ACCESS
	movlw	DSPEC2_720
	movwf	CURDSP2,ACCESS
	movlw	DSPD_720
	movwf	CURDSPD,ACCESS
	movlw	SpT_720
	movwf	CURSPT,ACCESS
	return

; ACCESS space: bank 0 addresses 0..5F (0..5F), bank 15 addresses 60..FF (F60..FFF)
; important: ANSELx locations are out of the access space (below abslote address F60)!

main:
	movlw	1
	movwf	DSK0CHGFLG,ACCESS		; disk changed on startup
	movwf	DSK1CHGFLG,ACCESS

	clrf	DSK0SPD,ACCESS
	clrf	DSK1SPD,ACCESS			; setting speed 500kB/s

	; setting default speeds
	call	Set144SpeedVals

	clrf	DEBUGN,ACCESS

	; set timter TMR1
	clrf	TMR1H,ACCESS
	clrf	TMR1L,ACCESS

	; turning timer1 on (driven by internal clock fosc/4)
	movlw	0x11					; timer 1 on, prescaler 1:2 (slower interrupt rate and thus
	movwf	T1CON,ACCESS			; longer time for drive motor to shut down
	movlw	0x01
	movwf	PIE1,ACCESS				; timer1 overflow interrupt enable
	movlw	0xC0
	movwf	INTCON,ACCESS			; enable interrupts globall

	; increasing running speed to 64MHz
	movlw	0xf0					; internal oscillator, 16MHz
	movwf	OSCCON,ACCESS			; PLL on, 32MHz
	movlw	0x5f
	movwf	OSCTUNE,ACCESS			; maximal frequency with PLL enabled (CONFIG1H sets INTIO67)

	; all ditigal pins setup
	movlb	0x0f
	clrf	ANSELA,BANKED
	clrf	ANSELB,BANKED
	clrf	ANSELC,BANKED
	clrf	ANSELD,BANKED
	clrf	ANSELE,BANKED			; all ports are digital outputs and inputs
	movlb	0

	movlw	0xFD					; RE1 as output, other pins as input
	movwf	TRISE,ACCESS
	movlw	0xFF
	movwf	TRISB,ACCESS
	movwf	TRISD,ACCESS			; ports A, B, D as input (W equals to 0xFF)
	clrf	TRISC,ACCESS			; port C as output
	clrf	TRISA,ACCESS			; port A as output

; DIAGNOSTIC PORT SETTING (port B as output)
	clrf	TRISB,ACCESS

	movwf	PORTA,ACCESS
	movwf	PORTB,ACCESS
	movwf	PORTC,ACCESS
	movwf	PORTD,ACCESS
	movwf	PORTE,ACCESS			; all ports A-D output high level (0xFF)

; DIAGNOSTIC PORT SETTING (port B outputs 0)
	clrf	PORTB,ACCESS

	; clear controller status flag
	clrf	CTRSTAT,ACCESS
	clrf	CMDRUN,ACCESS
	clrf	MOTATO,ACCESS
	clrf	MOTBTO,ACCESS
	clrf	DRINUM,ACCESS
	clrf	CURDRV,ACCESS
	clrf	CURHOSTDRV,ACCESS

LoopReset:
	movf	PORTE,W,ACCESS
	andlw	0xC					; RE3/RE2
	btfss	STATUS,Z,ACCESS
	goto	LoopReset			; loop until both are 0 (inactive)

	movlw	0xFF
	call	Wait_ms				; wait 255 ms

HostWaitMainLoop:
	bsf		INTCON,GIE,ACCESS		; enable interrupts
	call	WaitForHostWrite		; wait for host write. Interrupts are enabled, thus host write may be missed by the MC.
									; If MC misses command write, then host will next read 0xFF. If MC did not miss command write
									; it disables interrupts and returns command initiation byte 0xAA (or 0xAB, depending
									; on request)

	; process looks the following way:
	; 1. host ensures that MC is available/free through using loopback command
	; 2. host issues write command with command needed to port base+0. If MC was in the process of some command, it aborts
	;    the command and returns to HostWaitMainLoop
	; 3. host writes parameters to the port base+1 of MC. Host can write anything to port base+0 and this action aborts the sequence
	;    returning 0zAA and going to HostWaitMainLoop
	; 4. As soon as MC receives required set of parameters, it starts processing request. Depending on command MC returns command execution
	;    byte to the host. Host checks progress by reading from base port+1, if receives 0xFF this means that MC is not ready.
	; 5. As soon as host reads next valid value from port base+1, it performs additional data transfers (reads or writes) to the poer base+1
	; 6. When required number of bytes were transferred, host starts to receive status information.
	; 7. When all return data is read, MC returns to HostWaitMainLoop and waits for next command. Interrupts are enabled, and disk drive motor
	;    delay countdown is processed.

	; examination of the requested operation (WREG)
	; 0xAA = reset command progress and echo/looback (returns 0xAA to the host)
	; 0x00 = INIHRD, initialize FDC and FDDs (confirmation parameter 0xEC)
	; 0x01 = DSKIO read sector (drive#+head, cylinder, sector number; returns MC standard error data, see below)
	; 0x02 = DSKIO write sector (same as read sector)
	; 0x03 = DSKCHG check for disk change (drive#; returns 0xFF disk changed, 0x01 disk unchanged)
	; 0x04 = DSKFMT disk format (confirmation parameter 0xEC, and then drive#)
	; 0x05 = DRIVES return number of physical drives and currently selected drive (0-3)
	; 0x06 = VERSION returns MC version string
	; > 0x06 = reset request, return 0x55 to the host and restart sensing new request

	btfsc	STATUS,C,ACCESS			; A0 address line state during host request
	goto	HostWaitMainLoop		; access to port base+1 is not assumed this time

	bcf		INTCON,GIE,ACCESS		; disable interrupts

	movwf	HOSTCMD,ACCESS
	sublw	0xAA
	btfsc	STATUS,Z,ACCESS
	goto	NoCommandToRun

	movf	HOSTCMD,W,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	INITHRD
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	DSKIO
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	DSKIO
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	DSKCHG
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	DSKFMT
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	DRIVES
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	VERSION

NoCommandToRun:
	movlw	0xAA
	call	WaitForHostRead				; returns 0xAA to the host = "loopback", or "no command to run"
	goto	HostWaitMainLoop

; --- Give own version to the host
VERSION:
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	; and now return ID string hardcoded in MC's EPROM
	movlw	0x02				; +2 (not +1)
	movwf	FMTTMP3,ACCESS
IDLoop:
	movf	FMTTMP3,W,ACCESS
	addlw	LOW IDTable
	movwf	FMTTMP1,ACCESS
	movlw	HIGH IDTable
	btfsc	STATUS,C
	addlw	1
	movwf	PCLATH,ACCESS
	movlw	UPPER IDTable
	btfsc	STATUS,C
	addlw	1
	movwf	PCLATU,ACCESS
	incf	FMTTMP3,1,ACCESS
	incf	FMTTMP3,1,ACCESS	; +2 (not +1)
	movf	FMTTMP1,W,ACCESS
	call	IDTable
	movwf	FMTTMP2,ACCESS
	call	WaitForHostRead
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	HostWaitMainLoop			; access to port base+1 is not assumed this time, aborting MC command

	movf	FMTTMP2,W,ACCESS
	sublw	0xFF
	btfss	STATUS,Z
	goto	IDLoop
	goto	HostWaitMainLoop

IDTable:
	movwf	PCL,ACCESS
	retlw	"P"
	retlw	"I"
	retlw	"C"
	retlw	"1"
	retlw	"8"
	retlw	"F"
	retlw	"4"
	retlw	"6"
	retlw	"K"
	retlw	"2"
	retlw	"2"
	retlw	0x20
	retlw	"v"
	retlw	"."
	retlw	"2"
	retlw	"."
	retlw	"0"
	retlw	0xFF

; --- INIHRD host routine
INITHRD:
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	call	WaitForHostWrite			; read confirmation code
	btfss	STATUS,C,ACCESS				; address A0 is set to 1 (reset MC operation process)
	goto	NoCommandToRun				; in this case return 0xAA

	sublw	0xEC						; pass word to start initialization of the FDD hardware
	btfss	STATUS,Z,ACCESS
	goto	NoCommandToRun

	call	InitHardware

	movlw	0x80
	call	WaitForHostRead				; return command completion code to the host
	goto	HostWaitMainLoop

; --- DRIVES host routine
DRIVES:
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	movf	DRINUM,W
	call	WaitForHostRead				; return number of drives to host
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	movf	CURHOSTDRV,W,ACCESS
	call	WaitForHostRead				; return current drive selected (used for single physical drive config)
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	goto	HostWaitMainLoop

; --- DSKIO host routine
DSKIO:
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	call	WaitForHostWrite			; get first parameter: drive # (0=A:) and head (in FDC format)
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	movwf	DRVPRM,ACCESS

	call	WaitForHostWrite			; get second parameter: cylinder#
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	movwf	CYLPRM,ACCESS

	call	WaitForHostWrite			; get third parameter: sector# (track starts with sector 0x01)
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	movwf	SECPRM,ACCESS

	; after successfully getting third parameter MC command starts to execute. Host should monitor
	; return at the base port+1 to become 0xAE for data transfer or for 0xAF for status transfer

	; turn motor on
	movf	DRVPRM,W,ACCESS
	andlw	0x01
;	movwf	CURDRV					; drive number (0/1)
	call	CheckSingleDrive
	call	MotorOnSelect

	movlw	2
	movwf	SPDCOU,ACCESS			; trying two speeds

SpeedTryLoop:
	; take remembered value for speed for specific drive (0/1)
	movf	DSK0SPD,W,ACCESS
	btfsc	CURDRV,0,ACCESS
	movf	DSK1SPD,W,ACCESS

	btfss	WREG,0,ACCESS
	goto	Set144
	call	Set720SpeedVals
	goto	SpeedsSet
Set144:
	call	Set144SpeedVals

SpeedsSet:
	; set speeds for specific drive according to its remembered mode
	call	SetSpeeds

	movf	HOSTCMD,W,ACCESS
	decf	WREG,W,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	DoNotCheckWriteProtected

	; write operation - check for disk write protection issuing SENSE DRIVE STATUS command
	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x04
	call	AddToFDCQueue
	movf	DRVPRM,W,ACCESS
	call	AddToFDCQueue
	call	RunCommand

	movf	FDCERR,1,ACCESS
	btfss	STATUS,Z,ACCESS
	goto	DSKIORetUnkErr

NoSenseDSError:
	movf	CMDRLOC,W,ACCESS		; get value of ST3
	andlw	0x40
	btfsc	STATUS,Z,ACCESS
	goto	DoNotCheckWriteProtected

	; diskette is write protected
	clrf	WREG,ACCESS				; WRITE PROTECT error
	goto	DSKIORetStat

DoNotCheckWriteProtected:
	; check if disk was changed
	; if disk is not changed, then it does not matter if next seek command will really seek and set disk change bit
	; but if it's changed, then we seek to cylinder 1, then perform recalibrate, and check "disk change" bit again
	; if it's still set, it means there's no disk in drive - and we return with error
	call	ReadDiskChange
	btfss	TEMPREQ,7,ACCESS
	goto	NotChanged				; disk change bit set - proceed with seek and see if it will be still set

	movf	CYLPRM,W,ACCESS
	movwf	FMTPR1,ACCESS			; save current cylinder number

	movlw	0x01
	movwf	CYLPRM,ACCESS
	call	SeekCommand				; Cylinder 1
	btfsc	FDCERR,0,ACCESS
	goto	DSKIORetUnkErr			; recalibration error

	call	Recalibrate				; recalibrate
	btfsc	FDCERR,0,ACCESS
	goto	DSKIORetUnkErr			; recalibration error

	movf	FMTPR1,W,ACCESS
	movwf	CYLPRM,ACCESS

	call	ReadDiskChange
	btfsc	TEMPREQ,7,ACCESS
	goto	DSKIORetUnkErr			; disk change bit set after seek - no diskette present - return

NotChanged:
	; proceed with seek and data read/write
	call	SeekCommand				; nothing happens if heads are already positioned there
	movf	FDCERR,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	NoSeekError
DSKIORetSeekErr:
	movlw	0x06					; SEEK ERROR
	goto	DSKIORetStat

NoSeekError:
	; and try reading data
	call	NewFDCRequest			; clear MC FDC queue
	movf	HOSTCMD,W,ACCESS
	decf	WREG,W,ACCESS
	movlw	0x46					; HOSTSCMD=1 -- read operation 0x46
	btfss	STATUS,Z,ACCESS
	movlw	0x45					; HOSTSCMD=2 -- write operation 0x45
	call	AddToFDCQueue
	movf	DRVPRM,W,ACCESS			; head # + drive # (in FDC format)
	call	AddToFDCQueue
	movf	CYLPRM,W,ACCESS			; cylinder #
	call	AddToFDCQueue
	movf	DRVPRM,W,ACCESS
	rrncf	WREG,1,ACCESS
	rrncf	WREG,1,ACCESS
	andlw	1						; head #
	call	AddToFDCQueue
	movf	SECPRM,W,ACCESS			; sectorn #
	call	AddToFDCQueue
	movlw	0x02					; number of bytes (0x200 = 512)
	call	AddToFDCQueue
	movf	SECPRM,W,ACCESS			; last sector# on the track to transfer (the same as sector# to read/write)
	call	AddToFDCQueue
	movf	CURGPL1,W,ACCESS		; GAP1 default size
	call	AddToFDCQueue
	movlw	0xFF					; Data length (default 0xFF)
	call	AddToFDCQueue

	movf	HOSTCMD,W,ACCESS
	decf	WREG,W,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	DoNotLoadBufferFromHost	

	call	LoadDataToBufferFromHost		; host will write data to MC (FDD)
											; returns C=0 if aborted; C=1, Z=1 if ok; C=1, Z=0 for CRC error

	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	goto	HostWaitMainLoop		; cancel command completely
	btfss	STATUS,Z,ACCESS
	goto	DSKIORetUnkErr			; C=1, Z=0 -> CRC error, return "Unknown error" to host

DoNotLoadBufferFromHost:
	call	RunCommandWithData
	movf	FDCERR,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	NoAccessError

NextSpeed:
	; we are here because of error occured
	; change the speed, and try again
	decfsz	SPDCOU,1,ACCESS
	goto	ChangeSpeed
	goto	DSKIORetUnkErr

ChangeSpeed:
	movlw	0x01
	btfss	CURDRV,0,ACCESS
	goto	ChgDrv0
	; drive 1
	xorwf	DSK1SPD,1,ACCESS
	goto	SpeedTryLoop
ChgDrv0:
	xorwf	DSK0SPD,1,ACCESS
	goto	SpeedTryLoop

	; both speed attempts failed
DSKIORetUnkErr:
	movlw	0x0C					; "Unknown error"
	goto	DSKIORetStat

NoAccessError:
	call	DecodeError				; returns 0xCC if no error
	btfss	WREG,7,ACCESS			; bit 7 = 1 - no error occured
	goto	NextSpeed				; try another speed setting

	btfss	HOSTCMD,0,ACCESS		; bit 0 of command = 0 -> write op (10b) (do not wait host for reading data)
	goto	DSKIORetStat

	call	TransferDataToHostFromBuffer	; host will read data from MC (FDD)

	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	goto	HostWaitMainLoop		; cancel command completely
	btfss	STATUS,Z,ACCESS
	goto	DSKIORetUnkErr			; C=1, Z=0 -> CRC error, return "Unknown error" to host

DSKIOSuccRetStat:
	; save speed ID of successful r/w operation
	movf	SPDCOU,W,ACCESS
	btfss	CURDRV,0,ACCESS
	goto	SavDrv0
	; drive 1
	movwf	DSK1SPD,ACCESS
	goto	DSKIOExitSuccess
SavDrv0:
	movwf	DSK0SPD,ACCESS
DSKIOExitSuccess:
	movlw	0xCC

DSKIORetStat:
	; here host is trying to read return byte and then will read status byte

; DIAGNOSTIC DISPLAY - DSKIO return code
;movwf	PORTB

	movwf	SECPRM,ACCESS			; preserve return byte
	call	MotorOff

	movlw	0xAF
	call	WaitForHostRead			; returning 0xAF - result phase
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	movf	SECPRM,W,ACCESS			; restore return byte
	call	WaitForHostRead
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	goto	HostWaitMainLoop

; Routine seeks to specified cylinder and remembers 
; in: (DRVPRM) = drive number in FDC format, (CYLPRM) = cylinder to seek to
; updates HD*POS value to identify current head position of the drive
SeekCommand:
	call	ReadDiskChange
	btfsc	TEMPREQ,7,ACCESS
	call	DiskChangeFlagSet		; set disk change flag for specified drive for DSKCHG sub

	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x0F
	call	AddToFDCQueue
	movf	DRVPRM,W,ACCESS
	call	AddToFDCQueue
	movf	CYLPRM,W,ACCESS
	call	AddToFDCQueue
	call	RunCommand
	call	WaitSeekComplete
	return

; in: CURDRV = current drive ID (bit 0)
; out: none
DiskChangeFlagSet:
	btfss	CURDRV,0,ACCESS
	goto	FlagSetDrv0

	; drive 1 disk change flag setting
	bsf		DSK1CHGFLG,0,ACCESS
	return
FlagSetDrv0:
	bsf		DSK0CHGFLG,0,ACCESS
	return

; in: CURDRV = current drive ID (bit 0)
; out: none
DiskChangeFlagReset:
	btfss	CURDRV,0,ACCESS
	goto	FlagResetDrv0

	; drive 1 disk change flag setting
	clrf	DSK1CHGFLG,ACCESS
	return
FlagResetDrv0:
	clrf	DSK0CHGFLG,ACCESS
	return

; in: CURDRV = current drive ID (bit 0)
; out: W = flag (bit 0)
DiskChangeFlagGet:
	btfss	CURDRV,0,ACCESS
	goto	FlagGetDrv0

	; drive 1 disk change flag setting
	movf	DSK1CHGFLG,W,ACCESS
	return
FlagGetDrv0:
	movf	DSK0CHGFLG,W,ACCESS
	return


; -------------- Read sector data from host and write to MC buffer ------------------
; returns: buffer filled with data, C=0 (aborted writing to base port+0);
; C=1, Z=1 (ok); C=1, Z=0 (CRC error)
LoadDataToBufferFromHost:
	; initialize variables for data transfer
	clrf	RECLENL,ACCESS
	movlw	0x02
	movwf	RECLENH,ACCESS		; byte count in the record (512)

	movlw	0xAE
	call	WaitForHostRead		; signalling to host that 0x200 data byte stream will follow
	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	return							; access to port base+1 is not assumed this time, aborting MC command

	clrf	TMP,ACCESS			; clearing CRC byte
	movlw	0x01
	movwf	FSR0H,ACCESS
	clrf	FSR0L,ACCESS		; data buffer starts at absolute address 0x100

DIOWdata:
	call	WaitForHostWrite
	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	return							; access to port base+1 is not assumed this time, aborting MC command

	movwf	POSTINC0,ACCESS			; write data to MC buffer
	addwf	TMP,1,ACCESS			; add byte read from host to CRC

	decfsz	RECLENL,1,ACCESS
	goto	DIOWdata
	decf	RECLENH,1,ACCESS	; increase byte count
	bsf		STATUS,C,ACCESS		; set C flag (like access to base port+1)
	btfss	STATUS,Z,ACCESS
	goto	DIOWdata

	; now checking data CRC
	call	WaitForHostWrite	; read CRC from host
	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	return							; access to port base+1 is not assumed this time, aborting MC command

	subwf	TMP,1,ACCESS
	bsf		STATUS,C,ACCESS
	return							; return with Z flag valid: Z=1 then CRC matches, Z=0 then CRC does not match
; ************** Read sector data from host and write to MC buffer ******************


; -------------- Read sector data from MC buffer and send it to host ------------------
TransferDataToHostFromBuffer:
; input: buffer filled with data
; returns: C=0 (aborted writing to base port+0); C=1, Z=1 (ok); C=1, Z=0 (CRC error)

	; initialize variables for data transfer
	clrf	RECLENL,ACCESS
	movlw	0x02
	movwf	RECLENH,ACCESS			; byte count in the record (512)

	movlw	0xAE
	call	WaitForHostRead			; signalling to host that 0x200 data byte stream will need to be read
	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	return							; access to port base+1 is not assumed this time, aborting MC command

	clrf	TMP,ACCESS				; clearing CRC byte
	movlw	0x01
	movwf	FSR0H,ACCESS	
	clrf	FSR0L,ACCESS			; data buffer starts at absolute address 0x100

DIORdata:
	movf	POSTINC0,W,ACCESS		; read data from buffer and increase FSR0
	addwf	TMP,1,ACCESS			; add byte written to host to CRC
	call	WaitForHostRead
	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	return							; access to port base+1 is not assumed this time, aborting MC command

	decfsz	RECLENL,1,ACCESS
	goto	DIORdata
	decf	RECLENH,1,ACCESS		; decrease byte count
	bsf		STATUS,C,ACCESS			; set C flag (like access to base port+1)
	btfss	STATUS,Z,ACCESS
	goto	DIORdata

	; now checking data CRC
	call	WaitForHostWrite		; read CRC from host
	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	return							; access to port base+1 is not assumed this time, aborting MC command

	subwf	TMP,1,ACCESS
	bsf		STATUS,C,ACCESS
	return							; return with Z flag valid: Z=1 then CRC matches, Z=0 then CRC does not match
; ************** Read sector data from MC buffer and send it to host ******************

DecodeError:
	; Errors returned:
	; [0 = write protected] - not returned here, needs sense status command
	; [2 = not ready] - not returned here
	; 4 = crc error
	; 6 = seek error
	; 8 = record not found
	; 10 = write fault
	; 12 = other error
	; 0xCC = no error

	movf	FDCERR,1,ACCESS
	btfss	STATUS,Z,ACCESS
	goto	UnkError				; most probably timeout occured, sense data returned by FDC may not be valid

	; ST0
	btfsc	CMDRLOC,4,ACCESS		; first byte of the result data, ST0, bit 4 = "CHECK EQUIPMENT"
	goto	SeekError

	; ST1
	btfsc	CMDRLOC+1,5,ACCESS		; second byte of the result data, ST1, bit 5 = "DATA ERROR"
	goto	CrcError
	btfsc	CMDRLOC+1,2,ACCESS		; second byte of the result data, ST1, bit 2 = "NO DATA"
	goto	RecordNotFound
	btfsc	CMDRLOC+1,4,ACCESS		; second byte of the result data, ST1, bit 4 = "OVERRUN"
	goto	UnkError
	btfsc	CMDRLOC+1,1,ACCESS		; second byte of the result data, ST1, bit 1 = "NOT WRITABLE"
	goto	WriteFault
	btfsc	CMDRLOC+1,0,ACCESS		; second byte of the result data, ST1, bit 0 = "MISSING ADDRESS MARK"
	goto	RecordNotFound
;	btfsc	0x61,7					; second byte of the result data, ST1, bit 7 = "END OF CYLINDER"
;	goto	RecordNotFound

	; ST2
	btfsc	CMDRLOC+2,5,ACCESS		; second byte of the result data, ST1, bit 5 = "DATA ERROR"
	goto	CrcError
	btfsc	CMDRLOC+2,6,ACCESS		; second byte of the result data, ST1, bit 6 = "CONTROL MARK"
	goto	RecordNotFound
	btfsc	CMDRLOC+2,4,ACCESS		; second byte of the result data, ST1, bit 6 = "WRONG CYLINDER"
	goto	SeekError
	btfsc	CMDRLOC+2,1,ACCESS		; second byte of the result data, ST1, bit 6 = "BAD CYLINDER"
	goto	SeekError
	btfsc	CMDRLOC+2,4,ACCESS		; second byte of the result data, ST1, bit 6 = "NISSING ADDRESS MARK"
	goto	RecordNotFound

	movlw	0xCC					; no error
	return
UnkError:
	movlw	0x0C
	return
SeekError:
	movlw	0x06
	return
RecordNotFound:
	movlw	0x08
	return
CrcError:
	movlw	0x04
	return
WriteFault:
	movlw	0x0A
	return


DSKCHG:
	movlw	0xAB
	call	WaitForHostRead			; return 0xAB to the host - command has started
	btfss	STATUS,C,ACCESS			; address A0 is set to 1 (reset MC operation process)
	goto	NoCommandToRun			; in this case return 0xAA

	call	WaitForHostWrite		; and get first parameter: drive # (0=A:)
	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	goto	NoCommandToRun			; access to port base+1 is not assumed this time, aborting MC command

	andlw	0x01					; drive number
;	movwf	CURDRV
	call	CheckSingleDrive
	movwf	DRVPRM,ACCESS
	call	MotorOnSelect
	call	ReadDiskChange			; check FDD's disk change bit
	call	MotorOff

	btfsc	TEMPREQ,7,ACCESS
	goto	DSKChanged

	call	DiskChangeFlagGet
	btfsc	WREG,0,ACCESS
	goto	DSKChanged

	movlw	0x01					; disk unchanged
	goto	RetChg
DSKChanged:
	call	DiskChangeFlagReset
	; now clear the disk change bit by seeking to cylinder 10 first, and the cylinder 0
;	call	SoftReset
	call	MotorOnSelect
	movlw	0x01
	movwf	CYLPRM,ACCESS
	call	SeekCommand				; Cylinder 1
	clrf	CYLPRM,ACCESS
	call	SeekCommand				; Cylinder 0
	call	MotorOff
	movlw	0xFF					; disk changed

RetChg:
	movwf	COUVAL,ACCESS
	movlw	0xAF
	call	WaitForHostRead				; return 0xAF to indicate that results are ready
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	movf	COUVAL,W,ACCESS
	call	WaitForHostRead			; return value
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	goto	HostWaitMainLoop

; ---------------------------------------
; reads disk change attribute from LDCR - currently selected drive
; returns: TEMPREQ=LDCR byte read
ReadDiskChange:
	; read LDCR, bit 7
	movlw	0xFF
	movwf	PORTA,ACCESS
	clrf	PORTE,ACCESS			; activate /LDCR
	movlw	0xF0					; /LDOR=1, /RD=0, A=0, /CS=1: read FDC control register
	movwf	PORTC,ACCESS
	movlw	0xFF
	movwf	TRISA,ACCESS			; put port A into input mode
	nop
	nop
	movf	PORTA,W,ACCESS			; wait 62.5*3=187.5 (>90 which is minimal for read signal to be active)
	movwf	TEMPREQ,ACCESS

	movlw	0xF8
	movwf	PORTC,ACCESS			; release /RD
	clrf	TRISA,ACCESS			; set port A back to output mode
	movlw	0xFF
	movwf	PORTE,ACCESS			; release /LDCR

	; set disk change flag if diskchange bit is set
	btfsc	TEMPREQ,7,ACCESS
	call	DiskChangeFlagSet		; set disk change flag for DSKCHG sub
	return
; ---------------------------------------

DSKFMT:
	movlw	0xAB					; command accepted: now stage 2 to get first parameter from host
	call	WaitForHostRead			; return value
	btfss	STATUS,C,ACCESS			; address A0 is set to 1 (reset MC operation process)
	goto	NoCommandToRun			; in this case return 0xAA

	call	WaitForHostWrite
	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	goto	NoCommandToRun			; access to port base+1 is not assumed this time, aborting MC command

	movwf	DRVPRM,ACCESS
	andlw	0x01
;	movwf	CURDRV					; current drive
	call	CheckSingleDrive

	call	WaitForHostWrite		; get second: format parameter
	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	goto	NoCommandToRun			; access to port base+1 is not assumed this time, aborting MC command
	movwf	FMTPRM,ACCESS			; format parameter (1-standard 1.44, 2-compatible 1.44, , 3-compatible 720)

	call	WaitForHostWrite		; get third: confirmation code
	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	goto	NoCommandToRun			; access to port base+1 is not assumed this time, aborting MC command

	sublw	0xEC					; pass word
	btfss	STATUS,Z,ACCESS
	goto	NoCommandToRun

	movlw	0xAE
	call	WaitForHostRead			; return value to indicate that FORMAR command has started
	btfss	STATUS,C,ACCESS			; address A0 is set to 1 (reset MC operation process)
	goto	NoCommandToRun			; in this case return 0xAA

	; set drive speed
	call	Set144SpeedVals
	movf	FMTPRM,W,ACCESS
	sublw	0x03					; compatible 720kB
	btfsc	STATUS,Z,ACCESS
	call	Set720SpeedVals

	call	InitDrive				; initialize drive (CURDRV) to format disk in, and sets speeds

	clrf	CYLPRM,ACCESS			; init cylinder#=0
	clrf	SECPRM,ACCESS			; init head#=0

	; checking for diskette write protection - issuing SENSE DRIVE STATUS command
	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x04
	call	AddToFDCQueue
	movf	DRVPRM,W,ACCESS
	call	AddToFDCQueue
	call	RunCommand

	movf	FDCERR,1,ACCESS
	btfss	STATUS,Z,ACCESS
	goto	DSKIORetUnkErr			; "Unknown error"

	movf	CMDRLOC,W,ACCESS		; get value of ST3
	andlw	0x40
	btfsc	STATUS,Z,ACCESS
	goto	FormatLoop

	; diskette is write protected
	clrf	WREG,ACCESS				; WRITE PROTECT error
	goto	DSKIORetStat

FormatLoop:
	; proceed with seek to the specified cylinder
	call	SeekCommand
	movf	FDCERR,1,ACCESS
	btfss	STATUS,Z,ACCESS
	goto	DSKIORetSeekErr			; SEEK ERROR

FormatSeekOk:
	call	NewFDCRequest
	movlw	0x4D					; FORMAT TRACK command, MFM mode
	call	AddToFDCQueue
	movf	CURDRV,W,ACCESS
	iorwf	SECPRM,W,ACCESS			; OR with head number (in bit 2)
	call	AddToFDCQueue
	movlw	0x02					; Sector size = 512 bytes
	call	AddToFDCQueue
	movf	CURSPT,W,ACCESS			; Sectors per track (current value)
	call	AddToFDCQueue
	movlw	CURGPL2					; gap length 2 for formatting
	call	AddToFDCQueue
	movlw	0xE5					; Filler byte for data
	call	AddToFDCQueue

	movf	CYLPRM,W,ACCESS
	call	FillBufForFormat		; build tables to format specified track, should be run
									; AFTER format command buffer is filled with data (uses byte at offset 0x51)
									; it fills data for 18 consequtive sectors, even if we format 9 SpT

	call	RunCommandWithData
	movf	FDCERR,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	TrackFormatOk
FormatRetWriteFaultErr:
	movlw	0x0A					; WRITE FAULT
	goto	DSKIORetStat

TrackFormatOk:
	; next head
	movf	SECPRM,1,ACCESS
	btfss	STATUS,Z,ACCESS
	goto	NextCyl

	movlw	0x04
	addwf	SECPRM,1,ACCESS			; adds to the second FORMAT command parameter (bit 2++)
	goto	FormatSeekOk			; same track, side 1

NextCyl:
	clrf	SECPRM,ACCESS
	incf	CYLPRM,1,ACCESS
	movf	CYLPRM,W,ACCESS
	sublw	0x50					; track 79 is the last one
	btfss	STATUS,Z,ACCESS
	goto	FormatLoop

; --- Now writing control blocks: Boot sector and FAT
	; format options: 1=standard 1.44, 2=compatible 1.44, 3=compatible 720
	movf	FMTPRM,W,ACCESS			; starts with 1 (not 0)
	decfsz	WREG,1,ACCESS
	goto	Compat144_720Format

	; standard 1.44 format (1 sectors per cluster, 9 FAT sectors, GR8BIT stack at 0C382h)
	movlw	0x21
	movwf	FMTPR1,ACCESS			; control block sector count (33)
	movlw	0x20					; first FAT is at sector 1
	movwf	FMTPR2,ACCESS			; first FAT counter
	movlw	0x17					; seconf FAT is at sector 10 (=1+9)
	movwf	FMTPR3,ACCESS			; second FAT counter
	goto	ProceedWrite

Compat144_720Format:
	; we are here for options 2 and 3, former is compatible 1.44 and
	; latter is compatible 720, which have same control structures' allocation
	; boot=0, fat1=1, fat2=4, dir=7, data=14
	; in both cases (SpT=18 and SpT=9) structure sare located within cylinder 0

	; standard 1.44 format (4 sectors per cluster, 3 FAT sectors, 720K compatible stack location)
	movlw	0x0E
	movwf	FMTPR1,ACCESS			; control block sector count (33)
	movlw	0x0D					; first FAT is at sector 1
	movwf	FMTPR2,ACCESS			; first FAT counter
	movlw	0x0A					; second FAT is at sector 4 (=1+3)
	movwf	FMTPR3,ACCESS			; second FAT counter
	goto	ProceedWrite

ProceedWrite:
	movf	FMTPR1,W,ACCESS			; sector count (0x21 for standard and 0x0E for compatible)
	movwf	FMTTMP1,ACCESS			; current logical sector number in countdown mode
	clrf	CYLPRM,ACCESS			; current cylinder#
	movlw	0x01
	movwf	SECPRM,ACCESS			; currect sector number
	movf	CURDRV,W,ACCESS
	movwf	DRVPRM,ACCESS			; head bit in CURDRV = 0 (head 0)

WriteBootFat:
	; proceed with seek to the specified cylinder
	call	SeekCommand
	movf	FDCERR,1,ACCESS
	btfss	STATUS,Z,ACCESS
	goto	DSKIORetSeekErr			; SEEK ERROR

	call	ClearBufForFormat

	movf	FMTPR1,W,ACCESS			; sector count (0x21 for standard and 0x0E for compatible)
	subwf	FMTTMP1,W,ACCESS
	btfss	STATUS,Z,ACCESS
	goto	NotBootSector

	movlw	0x01
	movwf	FSR0H,ACCESS
	clrf	FSR0L,ACCESS			; set up indirect pointer to the buffer - 0x100

	movlw	0xb7
	movwf	FMTTMP2,ACCESS			; bytes in boot sector data

BootFill:
	movf	FMTTMP2,W,ACCESS
	sublw	0xb8
	movwf	FMTTMP3,ACCESS

	addwf	FMTTMP3,1,ACCESS		; bytepos*2 (16-bit words addressing by PIC18F46K22)

	movlw	HIGH BootSectTable
	btfsc	STATUS,C,ACCESS
	addlw	1
	movwf	PCLATH,ACCESS
	movlw	UPPER BootSectTable
	btfsc	STATUS,C,ACCESS
	addlw	1
	movwf	PCLATU,ACCESS			; add to the global indirect pointer

	movlw	LOW BootSectTable
	addwf	FMTTMP3,1,ACCESS
	btfsc	STATUS,C,ACCESS
	incf	PCLATH,1,ACCESS
	btfsc	STATUS,C,ACCESS
	incf	PCLATU,1,ACCESS
	movf	FMTTMP3,W,ACCESS		; and now add LOW table base to indirect pointer

	call	BootSectTable
	movwf	POSTINC0,ACCESS			; save boot sector byte to the stream

	decfsz	FMTTMP2,1,ACCESS
	goto	BootFill

	; now fille boot sector with proper disk configuration data
	movf	FMTPRM,W,ACCESS
	decfsz	WREG,1,ACCESS
	goto	CompatConfig

	movlb	1
	movlw	0x01			; Standard config: 1 sector per cluster, 224 dir entries, 9 sectors per FAT
	movwf	0x0D,BANKED		; cluster size
	movlw	0xE0
	movwf	0x11,BANKED		; dir entries
	movlw	0x09
	movwf	0x16,BANKED		; size of FAT
	movlb	0
	goto	NotFAT

CompatConfig:
	decfsz	WREG,1,ACCESS
	goto	Compat720

	movlb	1
	movlw	0x04			; Compatible config 1.44M: 4 sectors per cluster, 112 dir entries, 3 sectors per FAT
	movwf	0x0D,BANKED		; cluster size
	movlw	0x70
	movwf	0x11,BANKED		; dir entries
	movlw	0x03
	movwf	0x16,BANKED		; size of FAT
	movlb	0
	goto	NotFAT

Compat720:
	movlb	1
	movlw	0x02			; Compatible config 720K: 2 sectors per cluster, 112 dir entries, 3 sectors per FAT
	movwf	0x0D,BANKED		; cluster size
	movlw	0x70
	movwf	0x11,BANKED		; dir entries
	movlw	0x03
	movwf	0x16,BANKED		; size of FAT
	movlw	0xA0
	movwf	0x13,BANKED
	movlw	0x05
	movwf	0x14,BANKED		; sectors on disk (0x5A0)
	movlw	0xF9
	movwf	0x15,BANKED		; 720K media descriptor (0xF9)
	movlw	0x09
	movwf	0x18,BANKED		; number of sectors per track (0x9)
	movlb	0
	goto	NotFAT

NotBootSector:
	movf	FMTPR2,W,ACCESS			; first FAT pointer count (0x20 for standard and 0x0D for compatible)
	subwf	FMTTMP1,W,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	FATSector
	movf	FMTPR3,W,ACCESS			; second FAT pointer count (0x17 for standard and 0x0A for compatible)
	subwf	FMTTMP1,W,ACCESS
	btfss	STATUS,Z,ACCESS
	goto	NotFAT

FATSector:
	; select propert media descriptor for FAT copies
	movf	FMTPRM,W,ACCESS			; starts with 1 (not 0)
	decfsz	WREG,1,ACCESS
	goto	DescF9
	movlw	0xF0
	goto	EndDesc
DescF9:
	movlw	0xF9
EndDesc:
;	movlw	0xF0				; FAT descriptor
	movlb	1
	movwf	0x00,BANKED
	movlw	0xFF
	movwf	0x01,BANKED
	movwf	0x02,BANKED			; FAT "F0 FF FF ..."
	movlb	0

NotFAT:
	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x45
	call	AddToFDCQueue			; FDC write command
	movf	DRVPRM,W,ACCESS
	call	AddToFDCQueue
	movf	CYLPRM,W,ACCESS
	call	AddToFDCQueue
	movf	DRVPRM,W,ACCESS
	rrncf	WREG,1,ACCESS
	rrncf	WREG,1,ACCESS
	andlw	0x01
	call	AddToFDCQueue
	movf	SECPRM,W,ACCESS
	call	AddToFDCQueue
	movlw	0x02
	call	AddToFDCQueue
	movf	SECPRM,W,ACCESS			; last sector# on the track to transfer (the same as sector# to read)
	call	AddToFDCQueue
	movlw	CURGPL1					; GAP1 default size
	call	AddToFDCQueue
	movlw	0xFF					; Data length (default 0xFF)
	call	AddToFDCQueue
	call	RunCommandWithData

	movf	FDCERR,1,ACCESS
	btfss	STATUS,Z,ACCESS
	goto	FormatRetWriteFaultErr	; "WRITE FAULT"

	incf	SECPRM,1,ACCESS

	; setting max sector number +1
	movf	FMTPRM,W,ACCESS			; starts with 1 (not 0)
	decfsz	WREG,1,ACCESS
	goto	CompatSectors
	movlw	0x13					; 18 sectors per track
	goto	EndSectors
CompatSectors:
	movlw	0x0A					; 9 sectors per track
EndSectors:
;	movlw	0x13					; 18 sectors (+1 = overflow)
	subwf	SECPRM,W,ACCESS
	btfss	STATUS,Z,ACCESS
	goto	NoNewHead

	movlw	0x01
	movwf	SECPRM,ACCESS

	; next head here
	btfss	DRVPRM,2,ACCESS
	goto	NextHeadSameTrack

	; next track here
	bcf		DRVPRM,2,ACCESS
	incf	CYLPRM,1,ACCESS
	goto	NoNewHead

NextHeadSameTrack:
	bsf		DRVPRM,2,ACCESS

NoNewHead:
	decfsz	FMTTMP1,1,ACCESS
	goto	WriteBootFat

	; format ended successfully
	goto	DSKIOSuccRetStat

; ------------------- Boot sector data (called with index in W) --------------------
BootSectTable:			; disk configuration data and boot loader
	movwf	PCL
	; lookup table
	retlw	0xeb
	retlw	0xfe
	retlw	0x00
	retlw	0x47		; "GR8BIT10" format identifier
	retlw	0x52
	retlw	0x38
	retlw	0x42
	retlw	0x49
	retlw	0x54
	retlw	0x31
	retlw	0x30
	retlw	0x00		; bytes per sector (2)
	retlw	0x02
FormatChg1:	; offset 0x0D
	retlw	0x00		; cluster size in sectors (1) - 0x01 for standard and 0x04 for compatible formats
	retlw	0x01		; number of unused sectors (2)
	retlw	0x00
	retlw	0x02		; number of FATs (1)
FormatChg2:	; offset 0x11
	retlw	0x00		; number of directory entries (2) - 0xE0 for standard and 0x70 for compatible formats
	retlw	0x00
FormatChg3:	; offset 0x13
	retlw	0x40		; number of sectors per disk (2)
	retlw	0x0b
FormatChg4:	; offset 0x15
	retlw	0xf0		; media ID (1)
FormatChg5:	; offset 0x16
	retlw	0x00		; size of FAT in sectors (2) - 0x09 for standard and 0x03 for compatible formats
	retlw	0x00
FormatChg6:	; offset 0x18
	retlw	0x12		; number of sectors per track (2)
	retlw	0x00
	retlw	0x02		; number of sides (2)
	retlw	0x00
	retlw	0x00		; number of hidden sectors (2)
	retlw	0x00
	retlw	0xd0
	retlw	0xed
	retlw	0x53
	retlw	0x59
	retlw	0xc0
	retlw	0x32
	retlw	0xd0
	retlw	0xc0
	retlw	0x36
	retlw	0x56
	retlw	0x23
	retlw	0x36
	retlw	0xc0
	retlw	0x31
	retlw	0x1f
	retlw	0xf5
	retlw	0x11
	retlw	0xab
	retlw	0xc0
	retlw	0x0e
	retlw	0x0f
	retlw	0xcd
	retlw	0x7d
	retlw	0xf3
	retlw	0x3c
	retlw	0xca
	retlw	0x63
	retlw	0xc0
	retlw	0x11
	retlw	0x00
	retlw	0x01
	retlw	0x0e
	retlw	0x1a
	retlw	0xcd
	retlw	0x7d
	retlw	0xf3
	retlw	0x21
	retlw	0x01
	retlw	0x00
	retlw	0x22
	retlw	0xb9
	retlw	0xc0
	retlw	0x21
	retlw	0x00
	retlw	0x3f
	retlw	0x11
	retlw	0xab
	retlw	0xc0
	retlw	0x0e
	retlw	0x27
	retlw	0xcd
	retlw	0x7d
	retlw	0xf3
	retlw	0xc3
	retlw	0x00
	retlw	0x01
	retlw	0x58
	retlw	0xc0
	retlw	0xcd
	retlw	0x00
	retlw	0x00
	retlw	0x79
	retlw	0xe6
	retlw	0xfe
	retlw	0xfe
	retlw	0x02
	retlw	0xc2
	retlw	0x6a
	retlw	0xc0
	retlw	0x3a
	retlw	0xd0
	retlw	0xc0
	retlw	0xa7
	retlw	0xca
	retlw	0x22
	retlw	0x40
	retlw	0x11
	retlw	0x85
	retlw	0xc0
	retlw	0xcd
	retlw	0x77
	retlw	0xc0
	retlw	0x0e
	retlw	0x07
	retlw	0xcd
	retlw	0x7d
	retlw	0xf3
	retlw	0x18
	retlw	0xb4
	retlw	0x1a
	retlw	0xb7
	retlw	0xc8
	retlw	0xd5
	retlw	0x5f
	retlw	0x0e
	retlw	0x06
	retlw	0xcd
	retlw	0x7d
	retlw	0xf3
	retlw	0xd1
	retlw	0x13
	retlw	0x18
	retlw	0xf2
	retlw	0x42
	retlw	0x6f
	retlw	0x6f
	retlw	0x74
	retlw	0x20
	retlw	0x65
	retlw	0x72
	retlw	0x72
	retlw	0x6f
	retlw	0x72
	retlw	0x0d
	retlw	0x0a
	retlw	0x50
	retlw	0x72
	retlw	0x65
	retlw	0x73
	retlw	0x73
	retlw	0x20
	retlw	0x61
	retlw	0x6e
	retlw	0x79
	retlw	0x20
	retlw	0x6b
	retlw	0x65
	retlw	0x79
	retlw	0x20
	retlw	0x66
	retlw	0x6f
	retlw	0x72
	retlw	0x20
	retlw	0x72
	retlw	0x65
	retlw	0x74
	retlw	0x72
	retlw	0x79
	retlw	0x0d
	retlw	0x0a
	retlw	0x00
	retlw	0x00
	retlw	0x4d
	retlw	0x53
	retlw	0x58
	retlw	0x44
	retlw	0x4f
	retlw	0x53
	retlw	0x20
	retlw	0x20
	retlw	0x53
	retlw	0x59
	retlw	0x53
; ******************* Boot sector data (called with index in W) ********************

; ------------------- Initialize hardware (called from INIHRD) ---------------------
InitHardware:
	; --- Wait until system reset condition is cleared
	call	SoftReset

	clrf	DRINUM,ACCESS
	clrf	CURDRV,ACCESS
	call	InitDrive			; init drive 0 (A)
	btfss	STATUS,C,ACCESS
	incf	DRINUM,1,ACCESS		; now drive count = 1
	incf	CURDRV,1,ACCESS
	call	InitDrive			; init drive 1 (B)
	btfss	STATUS,C,ACCESS
	incf	DRINUM,1,ACCESS		; now drive count = 2

	clrf	CURDRV,ACCESS		; by default after hardware init drive 0 is selected
	clrf	CURHOSTDRV,ACCESS
	return
; ******************* Initialize hardware (called from INIHRD) *********************

SetSpeeds:
	; set channel speed
	movf	CURDSPD,W,ACCESS
	movwf	PORTA,ACCESS
	clrf	PORTE,ACCESS		; activate /LDCR

	movlw	0xE8				; /LDOR=1, /WR=0, A=0, /CS=1: write FDD controller operations register, base mode
	movwf	PORTC,ACCESS
	nop
	nop
	movlw	0xF8
	movwf	PORTC,ACCESS		; release /WR
	movlw	0xFF
	movwf	PORTA,ACCESS		; release data bus
	movwf	PORTE,ACCESS		; release /LDCR

; --- Specify (global attribute about disk drives' modes and times)
	call	NewFDCRequest
	movlw	0x03				; SPECIFY command
	call	AddToFDCQueue
	movf	CURDSP1,W,ACCESS
	call	AddToFDCQueue
	movf	CURDSP2,W,ACCESS
	call	AddToFDCQueue
	call	RunCommand
	return

; SoftReset:

; --- SOFT RESET OF FDD CONTROLLER ---
SoftReset:
	; Soft resetting FDD controller: LDOR=0/1, WR=1/0/1, 8 bit data bus
	clrf	PORTA,ACCESS		; SRST=0 (reset), DMAEN=0 (no DMA/IRQ)
	movlw	0x68				; /LDOR=0, /WR=0, A=0, /CS=1: write FDD controller operations register, base mode
	movwf	PORTC,ACCESS
	nop
	nop
	movlw	0xF8				; A=0, all others are 1
	movwf	PORTC,ACCESS
	movlw	0xFF
	movwf	PORTA,ACCESS		; setting data bus to high state

	movlw	0xFF
	call	Wait_ms				; wait 255 ms

	call	ReadMainStatRegFDC

	movlw	0x04				; release reset, DMAEN=0 (disabled)
	movwf	PORTA,ACCESS
	movwf	DORVAL,ACCESS
	movlw	0x68				; /LDOR=0, /WR=0, A=0, /CS=1: write FDD controller operations register, base mode
	movwf	PORTC,ACCESS
	nop
	nop
	movlw	0xF8
	movwf	PORTC,ACCESS
	movlw	0xFF
	movwf	PORTA,ACCESS
; --- RESET COMPLETE ---

	movlw	0xFF
	call	Wait_ms				; wait 255 ms

	call	ReadMainStatRegFDC
	call	SenseInterruptCmd

	movlw	0xFF
	movwf	TIMOUT1,ACCESS		; timeout: 255 loops
CheckAgain:
	call	ReadMainStatRegFDC	; read to get value and to get FDC stay in Base Mode
	btfsc	MAINSTAT,4,ACCESS
	goto	NotReady			; go there if bit 4 = 1 (command is executing)
	btfss	MAINSTAT,7,ACCESS
	goto	NotReady			; go there if bit 7 = 0 (FDC is not ready for data exchange)
	btfss	MAINSTAT,6,ACCESS
	goto	ReadyToWrite		; go there if bit 6 = 0 (ok to write to FDC)
	; FDC has something to say here
	movlw	0xFF
	movwf	PORTA,ACCESS		; port A - high level
	movlw	0xD1				; /FDDCS(0)+/RDI(0)+A0(1) = read data register
	movwf	PORTC,ACCESS
	nop
	nop
	movlw	0xF8
	movwf	PORTC,ACCESS
NotReady:
	decfsz	TIMOUT1,1,ACCESS
	goto	CheckAgain
	goto	SoftReset		; reset again after timeout

ReadyToWrite:
; --- Set up data rate - writing to CR

	call	SetSpeeds
	return


; ------------------------ Initialize drive ------------------------
; in:	CURDRV = current drive
; out:	RETERR = set if error occured at any of the init steps
InitDrive:
	; Turn motor on
	call	MotorOnSelect
	call	SetSpeeds

	; recalibrating drive
	call	Recalibrate
	btfsc	FDCERR,0,ACCESS
	goto	RetError

	; Seek to track 79
	call	NewFDCRequest
	movlw	0x0F				; SEEK command
	call	AddToFDCQueue
	movf	CURDRV,W,ACCESS
	call	AddToFDCQueue
	movlw	0x4F				; new cylinder number: 79
	call	AddToFDCQueue
	call	RunCommand			; no result phase
	call	WaitSeekComplete
	btfsc	FDCERR,0,ACCESS
	goto	RetError

; --- Seek to track 0, drive 0
	call	NewFDCRequest
	movlw	0x0F				; SEEK command
	call	AddToFDCQueue
	movf	CURDRV,W,ACCESS
	call	AddToFDCQueue
	movlw	0x00				; new cylinder number: 00
	call	AddToFDCQueue
	call	RunCommand			; no result phase
	call	WaitSeekComplete
	btfsc	FDCERR,0,ACCESS
	goto	RetError
	bcf		STATUS,C,ACCESS
	return

RetError:
	call	MotorOff	; start motor off countdown for current drive
	bsf		STATUS,C,ACCESS
	return
; ************************ Initialize drive ************************

; recalibrates CURDRV
; returns FDCERR!=0 on error
Recalibrate:
	; Recalibrate drive
	call	NewFDCRequest
	movlw	0x07				; RECALIBRATE command
	call	AddToFDCQueue
	movf	CURDRV,W,ACCESS
	call	AddToFDCQueue
	call	RunCommand
	call	WaitSeekComplete
	btfsc	FDCERR,0,ACCESS
	return

	; now we issue "SENSE DRIVE STATUS" command, and if track 0 signal (bit 4 of ST3)
	; is not set, drive is not present or defective
	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x04
	call	AddToFDCQueue			; Sense drive status FDC command
	movf	CURDRV,W,ACCESS
	call	AddToFDCQueue
	call	RunCommand

	btfsc	FDCERR,0,ACCESS
	return

	movlw	0xFF
	btfss	CMDRLOC,4,ACCESS		; bit 4 of ST3, "TRACK 0", is not set
	movwf	FDCERR,ACCESS
	return

; --------------- Run command --------------------
; in:	none
; out:	status bytes in bank 0 starting offset 0x40,
;		FDCRUNPOS=number of results read, FDCERR<>0 if error
RunCommandWithData:
	movlw	0xFF
	movwf	CMDRUN,ACCESS		; set "command is running" flag so that host receive status "busy"
	call	MotorOnSelect
	clrf	FDCERR,ACCESS
	call	CommandPhase
	call	ExecutePhase
	call	ResultPhase
	call	MotorOff
	clrf	CMDRUN,ACCESS
	return
; *************** Run command ********************

; --------------- Execute phase --------------------
ExecutePhase:
	clrf	TIMOUT2,ACCESS
	clrf	TIMOUT1,ACCESS		; long timeout = 0x10000 loops

	; wait for execution phase start
ExecWait:
	movlw	4
	call	Wait_us				; wait 4 us
	call	ReadMainStatRegFDC
	btfsc	MAINSTAT,5,ACCESS
	goto	ContinueExec

	; we are here if execution phase did not start (yet)
	; checking for RQM bit - if it's 1 but EXM is still 0, then
	; skip execution phase (probably error occured)
;	btfsc	MAINSTAT,7
;	goto	ExecError

	; and wait in loop
	decfsz	TIMOUT1,1,ACCESS
	goto	ExecWait
	decfsz	TIMOUT2,1,ACCESS
	goto	ExecWait
ExecError:
	movlw	0xFF
	movwf	FDCERR,ACCESS
	return

ContinueExec:
	clrf	RECLENL,ACCESS
	clrf	RECLENH,ACCESS			; byte count in the record

	clrf	TIMOUT2,ACCESS
	clrf	TIMOUT1,ACCESS			; long timeout = 0x10000 loops

	movlw	0x01
	movwf	FSR0H,ACCESS	
	clrf	FSR0L,ACCESS			; buffer start at 0x100

ExecuteSense:
	movlw	4
	call	Wait_us					; wait 4 us
	call	ReadMainStatRegFDC
	btfss	MAINSTAT,5,ACCESS		; EXM bit, equal to 0 when execution is complete
	return							; execute phase complete (with FDCERR preserved)

	btfsc	MAINSTAT,7,ACCESS
	goto	ExecRead

	decfsz	TIMOUT1,1,ACCESS
	goto	ExecuteSense
	decfsz	TIMOUT2,1,ACCESS
	goto	ExecuteSense
	; if we are here it means that timeout occured, but FDC is still not ready.
	; It may happen when diskette is unexpectedly removed - FDC will still wait for
	; something from FDD with bit 7 being 0 (not ready). The only way to clear this
	; condition is to perform soft reset of the FDC
	call	SoftReset
	goto	ExecError				; timeout error

ExecRead:
	btfss	MAINSTAT,6,ACCESS
	goto	ExecWrite				; go and write, otherwise read

	; data read from the FDC
	call	ReadDataRegFDC
	movwf	POSTINC0,ACCESS			; write data to buffer and increase FSR0
	goto	NextCount

ExecWrite:
	; data write to the FDC
	movf	POSTINC0,W,ACCESS		; read data from buffer and increase FSR0
	call	WriteDataRegFDC

NextCount:
	incfsz	RECLENL,1,ACCESS
	goto	ExecuteSense
	incf	RECLENH,1,ACCESS		; increase byte count
	goto	ExecuteSense
; *************** Execute phase ********************

; --------------- Run command --------------------
; in:	none
; out:	status bytes in bank 0 starting offset 0x40,
;		FDCRUNPOS=number of results read, FDCERR<>0 if error
RunCommand:
	clrf	FDCERR,ACCESS
	movlw	0xFF
	movwf	CMDRUN,ACCESS			; set "command is running" flag so that host receive status "busy"
	call	MotorOnSelect
	call	CommandPhase
	movlw	0x20
	call	Wait_us					; wait 32 microseconds
	call	ResultPhase
	call	MotorOff
	clrf	CMDRUN,ACCESS
	return
; *************** Run command ********************

; --------------- Reads data byte from the FDD controller --------------------
; in:	none
; out	data byte in W
ResultPhase:
	clrf	FSR0H,ACCESS			; indirect addressing bank 0
	movlw	CMDRLOC
	movwf	FSR0L,ACCESS			; low byte of the pointer
	clrf	FDCRUNPOS,ACCESS

	movlw	0xFF					; long timout! ; 0x08
	movwf	TIMOUT2,ACCESS
	clrf	TIMOUT1,ACCESS			; timeout = 0x800 loops
ReadLoop:
	call	ReadMainStatRegFDC
	btfsc	MAINSTAT,7,ACCESS		; 0x80, RQM
	goto	DataRegReady
	movlw	0x30
	call	Wait_us

	decfsz	TIMOUT1,1,ACCESS
	goto	ReadLoop
	decfsz	TIMOUT2,1,ACCESS
	goto	ReadLoop

	; timeout - exit
ReadErrorExit:
	movlw	0xFF
	movwf	FDCERR,ACCESS
	return

CheckCommandEnd:
	btfss	MAINSTAT,4,ACCESS		; 0x10, CB
	return							; result phase complete - return without error
	goto	ReadErrorExit

DataRegReady:
	btfss	MAINSTAT,6,ACCESS		; DIO - should be 1, i.e. data ready for reading
	goto	CheckCommandEnd

	call	ReadDataRegFDC
	movwf	POSTINC0,ACCESS			; write byte to the indirectly addressed location

	incf	FDCRUNPOS,1,ACCESS		; increment current pointer to the queue, increase pointer
	movf	FDCRUNPOS,W,ACCESS
	sublw	0x10
	btfss	STATUS,Z,ACCESS
	goto	ReadLoop
	goto	ReadErrorExit			; result buffer overrun (max 16 result bytes)
; *************** Reads data byte from the FDD controller ********************

; ----------------------- COMMAND PHASE ------------------------
; Writes queue contents to the FDC
CommandPhase:
	clrf	FSR0H,ACCESS			; indirect addressing bank 0
	movlw	CMDWLOC
	movwf	FSR0L,ACCESS			; low byte of pointer to the buffer
	clrf	FDCRUNPOS,ACCESS

	clrf	TIMOUT2,ACCESS
	clrf	TIMOUT1,ACCESS			; timeout = 0x800 loops

WriteLoop:
	call	ReadMainStatRegFDC
	btfsc	MAINSTAT,7,ACCESS		; 0x80, RQM
	goto	WriteRQMReady
CPhWait:
	movlw	0x30
	call	Wait_us
	decfsz	TIMOUT1,1,ACCESS
	goto	WriteLoop
	decfsz	TIMOUT2,1,ACCESS
	goto	WriteLoop
	; timeout - exit

WriteErrorExit:
	movlw	0xFF
	movwf	FDCERR,ACCESS
	return

WriteRQMReady:
	btfsc	MAINSTAT,6,ACCESS
	goto	CPhWait					;WriteErrorExit		; DIO=1 means controller has unexpectedly something to say
	movf	POSTINC0,W,ACCESS		; read byte from the indirectly addressed location and increase pointer
	call	WriteDataRegFDC

	incf	FDCRUNPOS,1,ACCESS		; increment current pointer to the queue
	movf	FDCRUNPOS,W,ACCESS
	subwf	FDCQPOS,W,ACCESS
	btfss	STATUS,Z,ACCESS
	goto	WriteLoop

	movlw	0x20
	call	Wait_us
	return						; FDCERR is preserved
; *********************** COMMAND PHASE ************************

; ------------------- Reads FDC data register -----------------------
; returns data byte in W
ReadDataRegFDC:
	movlw	0xFF
	movwf	PORTA,ACCESS		; port A - high level
	movlw	0xD1				; /FDDCS+/RDI+A0(1) = read data register
	movwf	PORTC,ACCESS
	movlw	0xFF
	movwf	TRISA,ACCESS
	nop
	movf	PORTA,W,ACCESS
	movwf	DATAVAL,ACCESS
	clrf	TRISA,ACCESS
	movlw	0xF8				; A=0, all others are 1
	movwf	PORTC,ACCESS
	movf	DATAVAL,W,ACCESS
	return
; ******************* Reads FDC data register ***********************

; ------------------- Writes to FDC data register -----------------------
; in:	W = data byte to write
WriteDataRegFDC:
	movwf	PORTA,ACCESS
	movlw	0xC9				; /FDDCS+/WRI+A0(1) = write data register
	movwf	PORTC,ACCESS
	nop
	nop
	movlw	0xF8				; A=0, all others are 1
	movwf	PORTC,ACCESS
	nop							; 125 ns delay to for FDC to latch and process byte
	return
; ******************* Writes to FDC data register ***********************

; --------------- Reads FDC master status register ----------------------
; returns status byte in W
ReadMainStatRegFDC:
;	movlw	0x30		; 24 us
;	call	Wait_us		; wait 24 us as per chip specification

	movlw	0xFF
	movwf	PORTA,ACCESS		; port A - high level
	movlw	0xD0				; /FDDCS+/RDI+A0 = read master register
	movwf	PORTC,ACCESS
	movlw	0xFF
	movwf	TRISA,ACCESS
	nop
	movf	PORTA,W,ACCESS
	movwf	MAINSTAT,ACCESS

	clrf	TRISA,ACCESS
	movlw	0xF8				; A=0, all others are 1
	movwf	PORTC,ACCESS
	movf	MAINSTAT,W,ACCESS
	return
; *************** Reads FDC master status register **********************

; -------------- Adds byte to the FDD command queue ---------------------
AddToFDCQueue:
	movwf	QTEMP,ACCESS
	clrf	FSR0H,ACCESS		; indirect addressing bank 0
	movf	FDCQPOS,W,ACCESS
	addlw	CMDWLOC
	movwf	FSR0L,ACCESS		; visible in all banks
	movf	QTEMP,W,ACCESS
	movwf	INDF0,ACCESS		; write to the indirectly addressed location
	incf	FDCQPOS,1,ACCESS
	return
; ************** Adds byte to the FDD command queue *********************

; ---------------- Clears FDD command/data queue ------------------------
NewFDCRequest:
	clrf	FDCQPOS,ACCESS
	return
; **************** Clears FDD command/data queue ************************

; ------------------------ SENSE INTERRUPT -------------------------------
SenseInterruptCmd:
	movlw	0x14
	call	Wait_ms				; wait 10 ms
	call	NewFDCRequest
	movlw	0x08				; SENSE INTERRUPT command
	call	AddToFDCQueue
	call	RunCommand
	return
; ************************ SENSE INTERRUPT *******************************

; ---------------- Wait millisecond, number of milliseconds in W -----------------
; W is preserved
Wait_ms:
	movwf	DELAY3,ACCESS
	clrf	DELAY1,ACCESS
Wait_ms2:
	movlw	0x14				; 0x0A on 32MHz
	movwf	DELAY2,ACCESS
Wait_ms1:
	decfsz	DELAY1,1,ACCESS
	goto	Wait_ms1
	decfsz	DELAY2,1,ACCESS
	goto	Wait_ms1
	decfsz	DELAY3,1,ACCESS
	goto	Wait_ms2
	return	
; **************** Wait millisecond, number of milliseconds in W *****************

; ---------------- Wait microsecond, number of microseconds in W -----------------
; 1+WREG*10 instruction cycles; for 64MHz one instruction takes 62.5ns, thus execution
; of 1 cycle is ~ equal to 625ns
Wait_us:
	movwf	DELAY1,ACCESS
Wait_us1:
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	decfsz	DELAY1,1,ACCESS
	goto	Wait_us1
	return
; **************** Wait microsecond, number of microseconds in W *****************

; ------------------ Turn motor on and clear turnoff counter ----------------
; in:	CURDRV = current drive #
MotorOnSelect:
;;;	bcf		INTCON,GIE			; disable interrupts
	clrf	TMP,ACCESS
	movf	DORVAL,W,ACCESS
	btfss	CURDRV,0,ACCESS
	goto	Drv0

	btfsc	DORVAL,5,ACCESS		; bit of drive motor running
	incf	TMP,1,ACCESS		; TMP to change to 1	

	clrf	MOTBTO,ACCESS		; no drive B timeout - keep it running
	andlw	0xFC				; mask drive select bits
	iorlw	0x21				; turn drive B motor on and select drive B
	goto	WriteDOR

Drv0:
	btfsc	DORVAL,4,ACCESS		; bit of drive motor running
	incf	TMP,1,ACCESS		; TMP to change to 1	

	clrf	MOTATO,ACCESS		; no drive A timeout - keep it running
	andlw	0xFC				; mask drive select bits
	iorlw	0x10				; turn drive A motor on and select drive A

WriteDOR:
	movwf	DORVAL,ACCESS		; save current DOR value
	movwf	PORTA,ACCESS
	movlw	0x68				; /LDOR=0, /WR=0, A=0, /CS=1: write FDD controller operations register, base mode
	movwf	PORTC,ACCESS
	nop
	nop
	movlw	0xF8
	movwf	PORTC,ACCESS
	movlw	0xFF
	movwf	PORTA,ACCESS

;;;	bsf		INTCON,GIE	; enable interrupts

	btfsc	TMP,0,ACCESS
	return						; skip drive spin up delay if TMP<>0 (selected drive motor was already running)

	movlw	0xFF
	call	Wait_ms
	movlw	0xFF
	call	Wait_ms				; wait 255*2=510 ms
	movlw	0xFF
	call	Wait_ms
	movlw	0xFF
	call	Wait_ms
	return
; ****************** Turn motor on and clear turoff counter ****************

; ------------------ Set up counter to turn motor off ----------------
; in:	CURDRV = current drive #
MotorOff:
	btfss	CURDRV,0,ACCESS
	goto	MotorCountDownDrv0

	movlw	0xFF
	movwf	MOTBTO,ACCESS		; drive B timeout set
	return

MotorCountDownDrv0:
	movlw	0xFF
	movwf	MOTATO,ACCESS		; drive A timeout set
	return
; ****************** Set up counter to turn motor off ****************

; ------------------ Wait for SEEK operation completes ----------------
WaitSeekComplete:
	call	SenseInterruptCmd
	btfsc	FDCERR,0,ACCESS
	return
	btfss	CMDRLOC,5,ACCESS
	goto	WaitSeekComplete	; loop while SE bit is 0 (seek is not complete)
	return
; ****************** Wait for SEEK operation completes ****************


; ------------------ Clear data buffer  ---------------------
; in:	none
ClearBufForFormat:
	clrf	RECLENL,ACCESS
	movlw	0x02
	movwf	RECLENH,ACCESS
	movlw	0x01
	movwf	FSR0H,ACCESS
	clrf	FSR0L,ACCESS			; buffer address is at 0x100

ClearFillLoop:
	clrf	POSTINC0,ACCESS
	decfsz	RECLENL,1,ACCESS
	goto	ClearFillLoop
	decfsz	RECLENH,1,ACCESS		; increase byte count
	goto	ClearFillLoop
	return
; ****************** Clear data buffer  *********************

; ------------------ Fill data buffer with data for format 1.44MiB diskette ------------
; in:	W = cylinder to format, 18 sectors per track
; (!!!) Important: sector interleaving is NOT implemented
FillBufForFormat:
	movwf	DATAVAL,ACCESS
	clrf	RECLENL,ACCESS
	clrf	RECLENH,ACCESS
	movlw	0x01
	movwf	TMP,ACCESS				; sector number
	movwf	FSR0H,ACCESS
	clrf	FSR0L,ACCESS			; buffer address = 0x100

FormatFillCycle:
	movf	DATAVAL,W,ACCESS		; cylinder # (track)
	movwf	POSTINC0,ACCESS

	rrncf	CMDWLOC+1,W,ACCESS
	rrncf	WREG,W,ACCESS
	andlw	0x01
	movwf	POSTINC0,ACCESS			; head
	movf	TMP,W,ACCESS
	movwf	POSTINC0				; sector#
	movlw	0x02
	movwf	POSTINC0,ACCESS			; sector size (0x200)
	incf	TMP,1,ACCESS			; increase number of the sector

	movf	TMP,W,ACCESS
	sublw	0x13					; let routine fill all 18 sector space
	btfss	STATUS,Z,ACCESS			; even if we format 9 sectors per track disk
	goto	FormatFillCycle
	return
; ****************** Fill data buffer with data for format 1.44MiB diskette ************

;============================= Microcontroller <-> Host communication =========================
; ---------------- Waits for host write --------------------
; in:	none
; out:	WREG=byte read from the host; CF=1 if A0 was 1 (port base+1)
; corrupt:	LATD, STATUS
WaitForHostWrite:
	movf	PORTE,W,ACCESS
	movf	PORTD,1,ACCESS			; read port D value and write it into the LATD
	btfss	WREG,3,ACCESS
	goto	WaitForHostWrite

	rrcf	WREG,W,ACCESS			; shift A0 address bit (bit 0) to C
	movf	LATD,W,ACCESS			; get previously latched input data to WREG
	return

; ---------------- Waits for host read --------------------
; in:	WREG=byte to return to host
; out:	CF=0 if was read from port base+0
; corrupt:	TRISD, LATD, STATUS
WaitForHostRead:
	movwf	PORTD,ACCESS			; write value to the port D output latch
									; PORTD is still functions as input
WaitWhilePICRDInactive:
	btfsc	PORTE,2,ACCESS
	goto	WaitWhilePICRDInactive	; Loop while /PICRD is active (skip this host read
									; cycle as its beginning is already missed)
WaitForPICRDAct:
	btfss	PORTE,2,ACCESS
	goto	WaitForPICRDAct			; Waiting for activation of /PICRD

	; /PICRD was activated - outputting port D
	clrf	TRISD,ACCESS

	; get address MC accessed with
	rrcf	PORTE,W,ACCESS			; "A0" (bit 0) -> CF

	; now watitng for /PICRD to deactivation
WaitForPICRDDeact:
	btfsc	PORTE,2,ACCESS
	goto	WaitForPICRDDeact		; Waiting for de-activation of /PICRD

	comf	TRISD,1,ACCESS			; !0 = 0xFF, now port D is input again (do not use decf as it affects C flag)
	movf	LATD,W,ACCESS			; restore initial value of WREG, does not affect C flag
	return
;============================= End of "Microcontroller <-> Host communication" =========================

;============================= Check for single disk drive configuration =========================
; needed to operate phantom drive 1
CheckSingleDrive:
; in:	drive number requested
	movwf	CURHOSTDRV,ACCESS
	movwf	CURDRV,ACCESS
	movf	DRINUM,W,ACCESS
	sublw	0x01					; is there only one drive?
	btfsc	STATUS,Z,ACCESS
	clrf	CURDRV,ACCESS			; current drive = drive 0 (while it can be phantom drive B: for host)
	return

;============================= Microcontroller interrupt handler =========================
; is used to turn off drives' motors. Is only executed when interrupts are enabled
int_handler:
	; if MOTxTO is not 0, decrement it and when it is zero stop the motor but not deselect drive
	movf	MOTATO,1,ACCESS			; check Z flag
	btfsc	STATUS,Z,ACCESS
	goto	StopDrvB				; MOTATO=0, do not count it down

	decfsz	MOTATO,1,ACCESS
	goto	StopDrvB				; counter did not reach 0 for drive A

	movf	DORVAL,W,ACCESS
	andlw	0xEF					; bit 4 reset
	call	WriteDOR

StopDrvB:
	movf	MOTBTO,1,ACCESS			; check Z flag
	btfsc	STATUS,Z,ACCESS
	goto	RetInt					; MOTBTO=0, do not count it down

	decfsz	MOTBTO,1,ACCESS
	goto	RetInt					; counter did not reach 0 for drive B

	movf	DORVAL,W,ACCESS
	andlw	0xDF					; bit 5 reset
	call	WriteDOR

RetInt:
	clrf	TMR1L,ACCESS
	clrf	TMR1H,ACCESS
	clrf	PIR1,ACCESS				; clear interrupt condition flag
	retfie	FAST					; return from interrupt and restore main registers
	END